#!/usr/bin/env python3

import os
import re
import sys


"""
2021-04-14, ed <irc.rizon.net>, MIT-licnenssed
https://ocv.me/dev/?hardlinken.py

regains disk space by deleting duplicate files,
replacing them with hardlinks, all by comparing
checksums in an mdw file as produced by hashwalk.py

so see also https://ocv.me/dev/?hashwalk.py

PLEASE NOTE that if you modify one of the files
then all the other hardlinks will also be modified
so only run this on stuff where this is fine
(for example c:\programfiles\adobe, like 20gib)

supports windows(ntfs), linux, tacos

if the mdw doesn't match your filesystem
(because you deleted some of the dupes) then
this will intentionally unceremoniously crash
"""


def undup(keep, drop):
    if os.path.samefile(keep, drop):
        return 0

    os.unlink(drop)
    os.link(keep, drop)
    return os.path.getsize(keep)


def main():
    hashfile = sys.argv[1]
    ptn = re.compile(rb'^([0-9]{6}) ([0-9-]{10} [0-9:]{8}) +([0-9]+) ([0-9a-f]{40}) (.*)')
    bexpect = 0
    nexpect = 0
    tab = {}
    with open(hashfile, 'rb') as f:
        for ln in f:
            m = ptn.match(ln)
            if not m:
                continue

            _, _, sz, ck, fp = m.groups()
            sz = int(sz)
            if sz < 16*1024:
                # smol, probably not worth
                continue

            fp = fp.rstrip(b'\r')
            hits = tab.get(ck)
            if hits:
                nexpect += 1
                bexpect += sz
                hits.append(fp)
            else:
                tab[ck] = [fp]

    tab = {k:v for k,v in tab.items() if len(v) > 1}
    n = sum([len(v) for v in tab.values()]) - len(tab)
    print(len(tab), n, nexpect)
    assert(n == nexpect)
    
    print(f"undup {nexpect:,} files, {bexpect:,} bytes")
    ndone = 0
    bdone = 0
    for x in tab.values():
        orig, dupes = x[0], x[1:]
        print(f"{n} {len(dupes)}x {orig}")
        for dupe in dupes:
            sz = undup(orig, dupe)
            n -= 1
            if sz:
                ndone += 1
                bdone += sz

    print(f"deleted {ndone:,} files, {bdone:,} bytes")
    if ndone < nexpect:
        nskip = nexpect - ndone
        bskip = bexpect - bdone
        print(f"skipped {nskip:,} files, {bskip:,} bytes  (already hard)")


if __name__ == '__main__':
    main()


# hashwalk.py x > c:\users\e\adobe.mdw
